/*
 * Copyright (C) 2008 Google Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.google.inject.spi;

import com.google.inject.Module;

/** @author jessewilson@google.com (Jesse Wilson) */
public class ElementApplyToTest extends ElementsTest {

  @Override
  protected void checkModule(Module module, ElementVisitor<?>... visitors) {
    // convert from module to elements and back
    super.checkModule(Elements.getModule(Elements.getElements(module)), visitors);
  }
}
