/*
 * Copyright (C) 2008 Google Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.google.inject.grapher;

import com.google.inject.Injector;
import com.google.inject.Key;
import java.io.IOException;
import java.util.Set;

/**
 * Guice injector grapher. Renders the guice dependency graph for an injector. It can render the
 * whole dependency graph or just transitive dependencies of a given set of nodes.
 *
 * @author phopkins@gmail.com (Pete Hopkins)
 * @since 4.0 (since 2.0 as a concrete class with a different API)
 */
public interface InjectorGrapher {

  /** Graphs the guice dependency graph for the given injector using default starting keys. */
  void graph(Injector injector) throws IOException;

  /**
   * Graphs the guice dependency graph for the given injector using the given starting keys and
   * their transitive dependencies.
   */
  void graph(Injector injector, Set<Key<?>> root) throws IOException;
}
