/*
 * Copyright (C) 2006 Google Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * Servlet API scopes, bindings and registration; this extension requires {@code guice-servlet.jar}.
 *
 * <p>Apply {@link com.google.inject.servlet.GuiceFilter} to any servlets which will use the servlet
 * scopes. Install {@link com.google.inject.servlet.ServletModule} into your {@link
 * com.google.inject.Injector} to install everything at once.
 */
package com.google.inject.servlet;
